package de.tudresden.inf.st.hybridpps.simulation.priority;

import de.tudresden.inf.st.hybridpps.simulation.SimulationProblem;
import de.tudresden.inf.st.hybridpps.simulation.model.SimulationJobMode;
import de.tudresden.inf.st.hybridpps.simulation.scheduling.AttributePool;
import de.tudresden.inf.st.hybridpps.simulation.scheduling.attributes.global.CombinedAttribute;
import de.tudresden.inf.st.hybridpps.simulation.scheduling.attributes.global.QueryableGlobalAttribute;
import de.tudresden.inf.st.hybridpps.simulation.scheduling.attributes.global.acc.Accumulator;
import de.tudresden.inf.st.hybridpps.simulation.scheduling.attributes.global.acc.AccumulatorPool;
import de.tudresden.inf.st.hybridpps.simulation.scheduling.attributes.modes.QueryableModeAttribute;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * The central class to represent the current state of the simulation which is used and queried as fact-base
 * by the meta-priority rule engine to decide on simulation parameters and strategies.
 * By default usage, only one SimulationState object is required at runtime which is updated on every iteration.
 */
public class SimulationState {

    private final Map<String, QueryableGlobalAttribute> globalAttributes;
    private final Map<String, QueryableModeAttribute> modeAttributes;
    private final Map<String, Accumulator> accumulators;

    private final SimulationProblem simulationProblem;
    private List<SimulationJobMode> executableModes;

    public SimulationState(SimulationProblem simulationProblem) {
        this.simulationProblem = simulationProblem;
        this.accumulators = AccumulatorPool.getAccumulators();
        globalAttributes = new HashMap<>();
        AttributePool.globalAttributes.forEach(globalAttribute ->
                globalAttributes.put(globalAttribute.getName(), globalAttribute));
        modeAttributes = new HashMap<>();
        AttributePool.modeAttributes.forEach(modeAttribute ->
                modeAttributes.put(modeAttribute.getName(), modeAttribute));
    }

    /**
     * Setter to update the current list of executable modes required for attribute calculation.
     * This method must be called on every simulation iteration BEFORE any query calls are made.
     *
     * @param executableModes modes of the current simulation iteration
     */
    public void update(List<SimulationJobMode> executableModes) {
        this.executableModes = executableModes;
    }

    /**
     * Calculates the value of a global attribute given by name from the current simulation experiment.
     * This method is only called by rules generated by the meta RuleSet and fired by the RuleEngine
     * <p>
     * CombinedAttributes (based on sum, mean, ..) are expected as: "accumulator_function:mode_attribute"
     *
     * @param attribute name of the attribute to calculate
     * @return attribute value
     */
    public double query(String attribute) {
        if (attribute.contains(":")) {
            //AccumulatedModeAttribute
            String[] parts = attribute.split(":");
            String accumulator = parts[0];
            String provider = parts[1];
            QueryableModeAttribute providerFunction = modeAttributes.get(provider);
            Accumulator accumulatorFunction = accumulators.get(accumulator);
            return CombinedAttribute.calculate(providerFunction, accumulatorFunction, executableModes, simulationProblem);
        } else {
            //PrimitiveGlobalAttribute
            double result = globalAttributes.get(attribute).getValue(executableModes, simulationProblem);
            return result;
        }
    }

}
