import pygad
import subprocess
import pathlib
import glob
import os
import json
import numpy as np
from argparse import ArgumentParser


attributes = [
    "duration",
    "direct_successor_count",
    "direct_successor_sum_min_processing_time",
    "direct_successor_sum_max_processing_time",
    "project_start_time",
    "project_finish_time",
    "resources",
    "setup_needed",
    "setup_time",
    "total_successor_count",
    "total_successor_sum_min_processing_time",
    "total_successor_sum_max_processing_time",
    "waiting_job_count",
    "waiting_time"
]

reduced_attributes = [
    "duration",
    "direct_successor_count",
    "resources",
    "setup_time",
    "waiting_time"
]

config_path = "config.json"


class SimulatorExecutor:
    """ Class used to execute the simulator. """
    
    def __init__(self, problem_instance, strategy, spr=False, deterministic=False):
        """
        :param problem_instance: path to the problem instance
        :param strategy: path to the strategy or name of the simple priority when spr=True
        :param spr: if True, use a simple priority rule for scheduling
        :param deterministic: if True, use deterministic instead of stochastic simulation
        """
        self.work_dir = pathlib.Path().resolve()
        self.problem_instance = problem_instance
        
        self.strategy = strategy
        self.spr = spr
        self.deterministic = deterministic

        self.solution_dir = "solutions"
        if not os.path.exists(self.solution_dir):
            os.makedirs(self.solution_dir)

        self.simulator_bat = os.path.join(self.work_dir.parent,
                                          "hybridpps.starter", "build", "install",
                                          "hybridpps.starter", "bin", "hybridpps.starter.bat")
        assert os.path.isfile(self.simulator_bat), \
            f"Bat file to start simulator wasn't found under path '{self.simulator_bat}'"

        self.problem_path = os.path.join(self.work_dir.parent, self.problem_instance)
        assert os.path.isfile(self.simulator_bat), \
            f"Problem instance wasn't found under path '{self.problem_path}'"

    def run_simulator(self):
        """ Executes the simulator.

        :return: content of the solution file generated by the simulator
        """
        command = self._create_command()
        completed_process = subprocess.run(command, capture_output=True)

        if completed_process.returncode != 0:
            print(completed_process.stderr)
        return self._read_solution_file()

    def _read_solution_file(self):
        list_of_files = glob.glob(self.solution_dir + '/*') # * means all if need specific format then *.csv
        latest_file = max(list_of_files, key=os.path.getctime)
        return json.load(open(latest_file))
            
    def _create_command(self):
        if self.spr:
            strategy = ["--spr", self.strategy]
        else:
            strategy = ["--cpr", os.path.join(self.work_dir, self.strategy)]

        command = [
            self.simulator_bat,
            self.problem_path,
            "--out", os.path.join(self.work_dir, "solutions")
        ]
        command += strategy
        if self.deterministic:
            command += ["-d"]
        return command


class GeneticCPRGenerator:
    
    def __init__(self, attributes, problem_instance, comparative_result=0):
        """
        :param cpr_generator:
        :param comparative_result: value to compare the result of the fitness function against, e.g. value of another
        rule
        """
        self.attributes = attributes

        self.work_dir = pathlib.Path().resolve()
        self.problem_instance = problem_instance

        self.comparative_result = comparative_result
        
        fitness_function = lambda x, y: self._fitness(x, y)
        num_genes = len(attributes)

        ga_params = self._load_ga_parameters()

        num_generations = ga_params["num_generations"]
        num_parents_mating = ga_params["num_parents_mating"]
        sol_per_pop = ga_params["sol_per_pop"]
        init_range_low = ga_params["init_range_low"]
        init_range_high = ga_params["init_range_high"]
        parent_selection_type = ga_params["parent_selection_type"]
        keep_parents = ga_params["keep_parents"]
        crossover_type = ga_params["crossover_type"]
        mutation_type = ga_params["mutation_type"]
        
        self.base_name = f"{num_genes}att_{num_generations}gen_{sol_per_pop}_pop"
        
        self.ga_instance = pygad.GA(fitness_func=fitness_function,
                                    num_generations=num_generations,
                                    num_parents_mating=num_parents_mating,
                                    sol_per_pop=sol_per_pop,
                                    num_genes=num_genes,
                                    init_range_low=init_range_low,
                                    init_range_high=init_range_high,
                                    parent_selection_type=parent_selection_type,
                                    keep_parents=keep_parents,
                                    crossover_type=crossover_type,
                                    mutation_type=mutation_type,
                                    save_solutions=True,
                                    on_generation=self._callback)
        
    def run(self):
        """ Starts the genetic algorithm.

        :return: None
        """
        self.ga_instance.run()
        
    def save_best_solution(self):
        """ Writes the the attributes that have given the best results to a .json file.

        :return: None
        """
        solution, _, _ = self.ga_instance.best_solution()
        name = os.path.basename(self.problem_instance).split(".")[0] + "_" + self.base_name + ".json"
        self.create_strategy_file(solution, name)

    def get_result(self, solution):
        """ Calls the simulator on a given set of attributes and returns the generated result.

        :param solution: list of attributes and weights
        :return: content of the solution file created by the simulator
        """
        strategy_path = self.create_strategy_file(solution)
        executor = SimulatorExecutor(self.problem_instance, strategy_path)
        return executor.run_simulator()

    def create_strategy_file(self, solution, name="data.json"):
        """ Generates a .json file from the given attributes so that they can be used as input for the simulator

        :param solution: list of attributes and weights
        :param name: name of the file to write the attributes to
        :return: path to the generated file which is equal to the name parameter
        """
        parameters = []
        for attr, par in zip(self.attributes, solution):
            parameters.append({"name": attr, "value": par})

        with open(name, 'w') as fp:
            json.dump(parameters, fp, indent=4)

        return name

    def _fitness(self, solution, solution_idx):
        print("Fitness Solution " + str(solution_idx))
        result = self.get_result(solution)

        result = result[0]["strategies"][0]["results"]
        return self.comparative_result - np.mean(result)

    def _load_ga_parameters(self):
        return json.load(open(config_path))

    @staticmethod
    def _callback(x):
        print("Finished Generation " + str(x.generations_completed))
        print("--------------------------------------------")


def main(args):
    problem = args.problem
    reduced = args.reduced
    attr = reduced_attributes if reduced else attributes

    ga = GeneticCPRGenerator(attr, problem)
    ga.run()
    ga.save_best_solution()

    solution, solution_fitness, solution_idx = ga.ga_instance.best_solution()
    print(f"Fitness value of the best solution = {solution_fitness}")

    ga.ga_instance.plot_fitness()
    ga.ga_instance.plot_genes()

    print("Resulting attribute weights: ")
    for attr, par in zip(ga.attributes, solution):
        print(attr + ": " + str(round(par, 2)))


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument("-p", "--problem", action="store", required=True,
                        help="relative path to the problem instance definition")
    parser.add_argument("-r", "--reduced", action="store_true",
                        help="if set, use only a reduced set of attributes for the generation")
    args = parser.parse_args()

    main(args)
